; Program: blinky
;  Author: Andrew Turnquist <andrew@turnquist.name>
; Version: 1.0
; Revision: $Id:$
;
; Description: Simple sample program for the asm1825 assembler.
;              The "Hello, World!" of microcontroller programs.
;
; Assemble with 'asm1825 -p blinky.asm'



	include pic16f1825.inc

	__config1	CONFIG_FCMEN_OFF & CONFIG_IESO_OFF & CONFIG_CLKOUTEN_OFF & CONFIG_BOREN_OFF & CONFIG_CPD_OFF & CONFIG_CP_OFF & CONFIG_MCLRE_OFF &CONFIG_PWRTE_ON & CONFIG_WDTE_OFF & CONFIG_FOSC_INTOSC
	__config2	CONFIG_LVP_OFF & CONFIG_STVREN_ON & CONFIG_PLLEN_OFF & CONFIG_WRT_OFF


	macro	clear	; Clear (zero) register
	banksel	$1
	clrf	$1
	endm

	macro	movlf
	movlw	$1	; Load literal into W
	banksel	$2
	movwf	$2	; Save W into register
	endm

	macro	brz	; Branch if not zero
	btfsc	STATUS,Z
	bra	$1
	endm


; Constants
DTIME	EQU	250		; Delays of 250ms


; Memory locations used

	; Define variables in common section.
	psect common,class=COMMON
delaytime:	db	0x00	; Counter for delay routine


	; Start assembling code into default section.
	psect default
	org	0x0000
_reset:	goto	start

	org	0x0010


; Delay 1ms @ 500kHz clock frequency
; 500kHz = 125,000 instructions/second or 125 instruction/ms
; Call/return each take 2 cycles, leaving 121 instructions
; movlf will generate 3 one cycle instructions, leaving 118 instructions
; decfsz+bra takes 3 cycles either way, leaving 116 instruction cycles.
; Thus, we initialize MSDELAY to 116 / 3, or 39.
;
; This is an example of using psects to encapsulate routines and
; permit easy relocation of the code.  We define a relocatable
; psect for our code and data, and the assembler will find a place
; for it.
; 
	psect msdelaydata,class=COMMON
msdelay:	db	0x00	; Counter for 1ms delay routine

	psect msdelaycode,class=CODE
delay1ms:
	movlf	39,msdelay	; Load initial value
msloop:
	decfsz	msdelay,f	; Decrement
	bra	msloop		;   if not zero, loop again
	return			;   else return


	; Return to the default program section.
	psect default

; Delay W ms (1-255)
delay:
	movwf	delaytime	; Load initial value
dloop:
	call delay1ms		; Delay 1ms
	decfsz	delaytime,f	; Decrement
	bra	dloop		;   if not zero, loop again
	return			;   else return


start:
	banksel	ANSELC
	clrf	ANSELC
	movlf	0b00111110,TRISC
	movlf	0b00111010,OSCCON
loop:
	banksel	PORTC		; Set RC0
	bsf	PORTC,RC0
	movlw	DTIME		; Delay 250ms for 2Hz flash rate
	call delay		; Delay
	banksel PORTC		; Clear RC0
	bcf	PORTC,RC0
	movlw	DTIME		; Delay 250ms
	call delay		; Delay
	bra loop		; Go back, Jack, and do it again.


