/*
 * dasm1825 - Disassembler for PIC16F1825 and others.
 *
 * $Id: dasm1825.c,v 1.2 2017-08-12 00:30:26-04 andrew Exp andrew $
 * 
 * Copyright (c) 2017 Andrew Turnquist
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <unistd.h>

#include "hexfile.h"
#include "lc.h"
#include "memmap.h"
#include "opcodes.h"
#include "util.h"

mmhandle chipmem;


void do_000(const unsigned short instr)
{
  int f;
  char c;
  unsigned short topbits = (instr & 0x3f00) >> 8;
  char d;

  d = (instr & 0x0080) ? 'f' : 'W';
  f = instr & 0x007f;

  if (instr == 0x0064) { printf("clrwdt\n"); return; }
  if (instr == 0x0000) { printf("nop\n");    return; }
  if (instr == 0x0062) { printf("option\n"); return; }
  if (instr == 0x0001) { printf("reset\n");  return; }
  if (instr == 0x0063) { printf("sleep\n");  return; }
  if ((instr & 0x3ff8) == 0x0060) {
    f = (instr & 0x0007);
    c = 'A' + (f-5);
    printf("tris\t%c\n", c);
    return;
  }
  if (instr == 0x0008) { printf("return\n"); return; }
  if (instr == 0x0009) { printf("retfie\n"); return; }
  if (instr == 0x000a) { printf("callw\n");  return; }
  if (instr == 0x000b) { printf("brw\n");    return; }
  if ((instr & 0x3fe0) == 0x0020) {
    f = (instr & 0x001f);
    printf("movlb\t%d\n", f);
    return;
  }
  if ((instr & 0x3ffc) == 0x0100) { printf("clrw\n"); return; }

  if ((topbits == 0x00) && (instr & 0x0080))
      { printf("movwf\t0x%02x\n", f); return; }
  if ((topbits == 0x01) && (instr & 0x0080))
      { printf("clrf\t0x%02x\n", f); return; }
  if (topbits == 0x02) { printf("subwf\t0x%02x,%c\n", f, d); return; }
  if (topbits == 0x03) { printf("decf\t0x%02x,%c\n", f, d);  return; }
  if (topbits == 0x04) { printf("iorwf\t0x%02x,%c\n", f, d); return; }
  if (topbits == 0x05) { printf("andwf\t0x%02x,%c\n", f, d); return; }
  if (topbits == 0x06) { printf("xorwf\t0x%02x,%c\n", f, d); return; }
  if (topbits == 0x07) { printf("addwf\t0x%02x,%c\n", f, d); return; }

  printf("???\t\t; 0x%04x\n", instr);
}


void do_001(const unsigned short instr)
{
  int f;
  unsigned short topbits = (instr & 0x3f00) >> 8;
  char d;

  d = (instr & 0x0080) ? 'f' : 'W';
  f = instr & 0x007f;

  if (topbits == 0x08) { printf("movf\t0x%02x,%c\n", f, d);   return; }
  if (topbits == 0x09) { printf("comf\t0x%02x,%c\n", f, d);   return; }
  if (topbits == 0x0a) { printf("incf\t0x%02x,%c\n", f, d);   return; }
  if (topbits == 0x0b) { printf("decfsz\t0x%02x,%c\n", f, d); return; }
  if (topbits == 0x0c) { printf("rrf\t0x%02x,%c\n", f, d);    return; }
  if (topbits == 0x0d) { printf("rlf\t0x%02x,%c\n", f, d);    return; }
  if (topbits == 0x0e) { printf("swapf\t0x%02x,%c\n", f, d);  return; }
  if (topbits == 0x0f) { printf("incfsz\t0x%02x,%c\n", f, d); return; }


  printf("??? ; 00 1\n");
}

// Bit clear, bit set
void do_010(const unsigned short instr)
{
  int f;
  char c;
  int b;

  b = (instr & 0x0380) >> 7;
  f = instr & 0x007f;
  c = (instr & 0x0400) ? 's' : 'c';

  printf("b%cf\t0x%02x,%d\n", c, f, b);
}


// BTFSC, BTFSS
void do_011(const unsigned short instr)
{
  int f;
  char c;
  int b;

  b = (instr & 0x0380) >> 7;
  f = instr & 0x007f;
  c = (instr & 0x0400) ? 's' : 'c';

  printf("btfs%c\t0x%02x,%d\n", c, f, b);
}


// CALL
void do_100(const unsigned short instr)
{
  int k = (instr & 0x07ff);

  printf("call\t0x%04x\n", k);
}


// GOTO
void do_101(const unsigned short instr)
{
  int k = (instr & 0x07ff);

  printf("goto\t0x%04x\n", k);
}


void do_110(const unsigned short instr)
{
  int f;
  // char c;
  unsigned short topbits = (instr & 0x3f00) >> 8;
  char d;
  int k9;
  int l;

  d = (instr & 0x0080) ? 'f' : 'W';
  f = instr & 0x007f;
  k9 = extendsign((instr & 0x01ff), 9);
  l = instr & 0x00ff;

  if ((topbits & 0x3e) == 0x32) { printf("bra\t%d\t; 0x%04x\n", k9, instr); return; }
  if (topbits == 0x34) { printf("retlw\t0x%02x\t; 0x%04x\n", l, instr); return; }
  if ((instr & 0x3f80) == 0x3180)
      { printf("movlp\t0x%02x\t; 0x%04x\n", (l & 0x7f), instr); return; }
  if (topbits == 0x30)
      { printf("movlw\t0x%02x\t; 0x%04x\n", l, instr); return; }
  if (topbits == 0x37)
      { printf("asrf\t0x%02x,%c\t; 0x%04x\n", f, d, instr); return; }
  if (topbits == 0x35)
      { printf("lslf\t0x%02x,%c\t; 0x%04x\n", f, d, instr); return; }
  if (topbits == 0x36)
      { printf("lsrf\t0x%02x,%c\t; 0x%04x\n", f, d, instr); return; }
  if ((topbits == 0x31) && !(instr & 0x0080)) {
    d = (instr & 0x0040) ? '1' : '0';
    k9 = extendsign(instr & 0x003f, 6);
    printf("addfsr%c\t%d\t; 0x%04x\n", d, k9, instr);
    return;
  }

  printf("???\t\t; 0x%04x\n", instr);
}


void do_111(const unsigned short instr)
{
  int f;
  unsigned short topbits = (instr & 0x3f00) >> 8;
  char d;
  int l;

  d = (instr & 0x0080) ? 'f' : 'W';
  f = instr & 0x007f;
  l = instr & 0x00ff;

  if (topbits == 0x38)
      { printf("iorlw\t0x%02x\t; 0x%04x\n", l, instr); return; }
  if (topbits == 0x39)
      { printf("andlw\t0x%02x\t; 0x%04x\n", l, instr); return; }
  if (topbits == 0x3a)
      { printf("xorlw\t0x%02x\t; 0x%04x\n", l, instr); return; }
  if (topbits == 0x3b)
      { printf("subwfb\t0x%02x,%c\t; 0x%04x\n", f, d, instr); return; }
  if (topbits == 0x3c)
      { printf("sublw\t0x%02x\t; 0x%04x\n", l, instr); return; }
  if (topbits == 0x3d)
      { printf("addwfc\t0x%02x,%c\t; 0x%04x\n", f, d, instr); return; }
  if (topbits == 0x3e)
      { printf("addlw\t0x%02x\t; 0x%04x\n", l, instr); return; }

  printf("???\t\t; 0x%04x\n", instr);
}


void
getroot(char *r, const char *f)
{
  int i;

  strncpy(r, f, 1023);
  i = strlen(r);
  while ((r[i] != '.') && (i > 0)) i--;
  if (r[i] == '.') r[i] = '\0';
}


int main(int argc, char * argv[])
{
  int opt;
  int lastlc = 0;	// Last location with valid data
  unsigned short instr;	// Read instruction
  char *hex = NULL;
  char fnroot[MAXPATHLEN];

  lc = 0;
  pass = 0;
  while ((opt = getopt(argc, argv, "qv")) != -1) {
    switch (opt) {
      case 'q':
        msg_level--;
        break;
      case 'v':
        msg_level++;
        break;
      default:
        fprintf(stderr, "Usage: %s [ -qv ] <source>\n", argv[0]);
        exit(1);
    }
  }
  if (argv[optind] == NULL) {
    fprintf(stderr, "Usage: %s [ -qv ] <source>\n", argv[0]);
    exit(1);
  }
  hex = argv[optind];
  getroot(fnroot, hex);
  message(MSG_INFO, "Using root name '%s'", fnroot);

  chipmem = mmInit(8192, 0x0b, 256, 0x8000);
  mmLoadHexfile(chipmem, hex);

  printf("%04x:\tORG\t0x%04x\n", lc, lc);
  for (lc=0; lc<8192; lc++) {
    if (mmHasData(chipmem, lc)) {
      instr = mmRead(chipmem, lc);
      if (instr & 0xc000) continue;
      if (lc && (lastlc < (lc-1))) {
        printf("\n%04x:\tORG\t0x%04x\n", lc, lc);
      }
      printf("%04x:\t", lc);
      switch ((instr & 0x3800) >> 11) {
        case 0:
          do_000(instr);
          break;
        case 1:
          do_001(instr);
          break;
        case 2:
          do_010(instr);
          break;
        case 3:
          do_011(instr);
          break;
        case 4:
          do_100(instr);
          break;
        case 5:
          do_101(instr);
          break;
        case 6:
          do_110(instr);
          break;
        case 7:
          do_111(instr);
          break;
        default:
          break;
      }  // switch

      lastlc = lc;
    }  // if has data
  }


  mmFree(chipmem);
  exit(0);
}


