/*
 * asm1825 - Assembler for PIC16F1825 and others.
 *
 * $Id: asm1825.c,v 1.2 2017-08-12 00:30:26-04 andrew Exp andrew $
 *
 * Copyright (c) 2017 Andrew Turnquist
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>
#include <unistd.h>

#include "memmap.h"
#include "opcodes.h"
#include "rc.h"
#include "util.h"
#include "processor.h"


struct procspec proc;
mmhandle chipmem;


/*
 * procInit -- Initialize 'proc' to sane defaults (for 16F1825).
 */

void procInit(void)
{
  proc.codesize = 8192;
  proc.ramsize = 1024;
  proc.eesize = 256;
}


void procDoDirective(char *param)
{
  rc *procconf;
  int banks;
  char pstr[MAXPATHLEN];

  /* Processor directive only handled during first pass */
  if (pass) return;

  // Try current directory first.
  snprintf(pstr, sizeof(pstr), "%s.pdat", param);
  if ((procconf = rcLoadFile(pstr)) == NULL) {
    if (param[0] != '/') {
      snprintf(pstr, sizeof(pstr), "%s/%s.pdat", PROCINCLUDE_PATH, param);
      if ((procconf = rcLoadFile(pstr)) == NULL) {
        message(MSG_ERROR, "processor file '%s' not found.", param);
        exit(1);
      }
    } else {
      message(MSG_ERROR, "processor file '%s' not found.", param);
      exit(1);
    }
  }

  message(MSG_DEBUG, "processor: procconf is%s null", procconf ? " not" : "");
  message(MSG_DEBUG, "processor: (rc) codesize=%s", rcValue(procconf, "codesize"));
  if (rcValue(procconf, "codesize"))
      proc.codesize = strtol(rcValue(procconf, "codesize"), NULL, 0);
  if (rcValue(procconf, "ramsize"))
      proc.ramsize = strtol(rcValue(procconf, "ramsize"), NULL, 0);
  if (rcValue(procconf, "eesize"))
      proc.eesize = strtol(rcValue(procconf, "eesize"), NULL, 0);

  rcDestroy(procconf);

  message(MSG_INFO, "processor: %d words PM", proc.codesize);

  message(MSG_NOTICE, "Reconfiguring memory map for %s", param);
  mmFree(chipmem);
  chipmem = mmInit(proc.codesize, 0x0b, proc.eesize, 0x8000);

  /* Calculate number of banks */
  banks = (proc.ramsize - 16) / 80;
  /* Add 1 if there's a partial bank at end */
  if (((banks * 80) + 16) != proc.ramsize) banks++;
  psSetBanks(banks);
}

